/**
 * The MIT License (MIT)
 *
 * Copyright (c) 2014 UVC Ingenieure http://uvc.de/
 * Author: Max Holtzberg <mh@uvc.de>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include <avr/io.h>
#include <stdio.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/sleep.h>

#include "uart.h"

#ifndef F_CPU
   #warning "F_CPU not defined global, using local F_CPU from uart.c!"   
   #define F_CPU  8000000
#endif

#define NOP __asm__ __volatile__  ( "nop\n" )

#define FALSE 0
#define TRUE ~FALSE

#define RXBUFFERSIZE 64
#define TXBUFFERSIZE 64

volatile uint8_t rxhead, rxtail;
volatile char rxfifo[RXBUFFERSIZE];

volatile uint8_t txhead = 0, txtail = 0;
volatile int8_t txflag = 0;
volatile char txfifo[TXBUFFERSIZE];


ISR(USART_RXC_vect)
{
   uint8_t c, t;
   if (UCSRA & (1 << RXC)) {
      c = UDR;
      t = rxhead + 1;
      if (t >= RXBUFFERSIZE) {
         t=0;
      }
      if (t != rxtail) {
         rxfifo[rxhead] = c;
         rxhead = t;
      }
   }
}

ISR(USART_TXC_vect)
{   
   if (txhead != txtail) {
      UDR = txfifo[txtail];
      txtail++;
      if (txtail >= TXBUFFERSIZE)
         txtail=0;
   } else {
      UCSRB &= ~(1 << TXCIE);
      txflag=0;
   }  
}

int uart_init( uint32_t baudrate, uint8_t parity, uint8_t databits, uint8_t stopbits) {

    uint8_t ucsrc, csz;
	uint16_t ubrr;

    ubrr = ((long) F_CPU / (baudrate * 16L) - 1);

    csz = databits - 5;
    if (databits==9) {
       csz = 7;
    }
    ucsrc = (csz << 1) & 0x06;
    switch (parity) {
       case PARITY_NONE:
            break;
       case PARITY_SPACE:
               ucsrc |= (1 << UPM0);
            break;
       case PARITY_EVEN:
              ucsrc |= (1 << UPM1);
            break;
       case PARITY_ODD:
              ucsrc |= (1 << UPM1) | (1 << UPM0);
            break;
    }

    UCSRC = ucsrc;

    UCSRB = _BV(TXEN) | _BV(RXEN) | _BV(RXCIE);
    UCSRC = ucsrc;
	  
    UBRRH = ubrr >> 8;
    UBRRL = ubrr & 0xff; 	  

    rxhead = 0;
    rxtail = 0;
    txhead = 0;
    txtail = 0;
    txflag = 0;

    return 0;
}


int uart_putchar (char c) {
   uint8_t t;
   int8_t rc = 0;


   if (!txflag) {
      UDR = c;
      txflag = 1;
      UCSRB |= (1 << TXCIE);   

      rc = TRUE;
   } else {
      cli();
      t = txhead + 1;
      if (t >= TXBUFFERSIZE) {
         t=0;
      }
      if (t != txtail) {
         txfifo[txhead] = c;
         txhead = t;
         rc = TRUE;
      }
      sei();
   }

   return (rc);
}

int uart_puts (const char *s) {
   while (*s) {
      if (uart_putchar(*s)) {
         s++;
      }

   };
   return 0;
}

int uart_getchar (void) {
    uint8_t c;
    if (rxtail != rxhead) {
       c = rxfifo[rxtail];
       rxtail++;
       if (rxtail >= RXBUFFERSIZE) {
          rxtail = 0;
       }
       return c;
    } else {
       return -1;
    }
}
